#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "computation_graph.h"

#define GRAPH_MAX_DEPTH 20
#define GRAPH_MAX_INPUTS 20

// Array to store input values for passing to the execute function of each node
static double exec_input_vals[GRAPH_MAX_INPUTS];

struct computation_graph *create_graph() {
    struct computation_graph *the_graph = malloc(sizeof(struct computation_graph));
    if (!the_graph) {return NULL;}
    the_graph->nodes = malloc(sizeof(struct graph_node));
    if (!the_graph->nodes) { return NULL; }
    the_graph->n_nodes = 0;
    the_graph->size = 1;
    return the_graph;
}

static void reset_node_rec(struct computation_graph* graph, int node_id, int depth) {
    if (depth > GRAPH_MAX_DEPTH) {
        return;
    }
    struct graph_node* node = &graph->nodes[node_id];
    // Don't reset nodes that are already connected to something else
    // Don't reset nodes that have already been reset or discovered
    if (node->n_children != 1 || node->processed_state != UNPROCESSED) {
        return;
    }
    node->processed_state = DISCOVERED;
    int input_id;
    for (input_id = 0; input_id < node->type->n_inputs; input_id++) {
        int src_node_id = node->input_srcs[input_id].controller_id;
        if (src_node_id != -1) {
            reset_node_rec(graph, src_node_id, depth + 1);
        }
    }
    // Reset this node
    if (node->type->reset != NULL) {
        node->type->reset(node->state);
    }
    node->processed_state = PROCESSED;
}

int reset_node(struct computation_graph* graph, int node_id) {
    if (node_id >= graph->n_nodes) {
        return -1;
    }
    int i;
    for (i = 0; i < graph->n_nodes; i++) {
        graph->nodes[i].processed_state = UNPROCESSED;
    }
    reset_node_rec(graph, node_id, 0);
    return 0;
}

int graph_set_source(struct computation_graph *graph,
                     int dest_node_id, int dest_input, int src_node_id, int src_output) {
    if (dest_node_id >= graph->n_nodes || src_node_id >= graph->n_nodes) {
        return -1;
    }
    struct graph_node *dest_node = &graph->nodes[dest_node_id];
    struct graph_node *src_node = &graph->nodes[src_node_id];
    if (dest_input >= dest_node->type->n_inputs || src_output >= src_node->type->n_outputs) {
        return -1;
    }

    // If a previous source exists, remove one from its children count
    int prev_src_id = dest_node->input_srcs[dest_input].controller_id;
    if (prev_src_id != -1) {
        graph->nodes[prev_src_id].n_children -= 1;
    }
    src_node->n_children += 1; // Count destination node as a child
    dest_node->input_srcs[dest_input].controller_id = src_node_id;
    dest_node->input_srcs[dest_input].controller_output = src_output;
    reset_node(graph, src_node_id);
    return 0;
}

int graph_add_node(struct computation_graph *graph,
                   const char* name,
                   const struct graph_node_type *type,
                   void *state) {
    assert(type->n_inputs <= GRAPH_MAX_INPUTS);
    int new_id = graph->n_nodes;
    if (new_id >= graph->size) {
        int new_capacity = graph->n_nodes == 0 ? 1 : graph->n_nodes * 2;
        struct graph_node *node_arr = realloc(graph->nodes, sizeof(struct graph_node) * new_capacity);
        if (!node_arr) {
            return -1;
        }
        graph->size = new_capacity;
        graph->nodes = node_arr;
    }
    struct graph_node *new_node = &graph->nodes[new_id];
    new_node->name = strdup(name);
    new_node->type = type;
    new_node->state = state;
    new_node->n_children = 0;
    new_node->output_values = malloc(type->n_outputs * sizeof(double));
    new_node->param_values = malloc(type->n_params * sizeof(double));
    new_node->input_srcs = malloc(type->n_inputs * sizeof(struct input_type));
    // Check that malloc succeeded in every case which memory was requested
    if ((type->n_outputs && !new_node->output_values) ||
        (type->n_params && !new_node->param_values) ||
        (type->n_inputs && !new_node->input_srcs)) {
        return -1;
    }
    int i;
    for (i = 0; i < type->n_inputs; i++) {
        new_node->input_srcs[i].controller_id = -1;
    }
    graph->n_nodes += 1;
    // Reset block upon creation
    if (new_node->type->reset != NULL) {
        new_node->type->reset(new_node->state);
    }
    return new_id;
}

int graph_set_param_val(struct computation_graph *graph, int node_id, int param_id, double value) {
    if (node_id >= graph->n_nodes || param_id >= graph->nodes[node_id].type->n_params) {
        return -1;
    }
    graph->nodes[node_id].param_values[param_id] = value;
    return 0;
}

double graph_get_output(const struct computation_graph *graph, int node_id, int output_id) {
    if (node_id >= graph->n_nodes || output_id >= graph->nodes[node_id].type->n_outputs) {
        return 0;
    }
    return graph->nodes[node_id].output_values[output_id];
}

void graph_compute_node_rec(struct computation_graph *graph, int node_id, int depth) {
    if (depth >= GRAPH_MAX_DEPTH) {
        assert(1 == 0); // TODO :Xil_Assert false
        return;
    }
    if (node_id >= graph->n_nodes) {
        return;
    }
    struct graph_node *node = &graph->nodes[node_id];
    if (node->processed_state != UNPROCESSED) {
        return;
    }
    node->processed_state = DISCOVERED;
    int input_id;
    for (input_id = 0; input_id < node->type->n_inputs; input_id++) {
        int src_cntl_id = node->input_srcs[input_id].controller_id;
        if (src_cntl_id != -1) {
            graph_compute_node_rec(graph, src_cntl_id, depth + 1);
        }
    }
    // Populate the exec_input_vals array for computation
    for (input_id = 0; input_id < node->type->n_inputs; input_id++) {
        int src_cntl_id = node->input_srcs[input_id].controller_id;
        int src_output_id = node->input_srcs[input_id].controller_output;
        if (src_cntl_id != -1) {
            exec_input_vals[input_id] = graph->nodes[src_cntl_id].output_values[src_output_id];
        }
        else {
        	// Set input value to 0 if not connected
        	exec_input_vals[input_id] = 0;
        }
    }
    if (node->type->execute != NULL) {
        (*node->type->execute)(node->state, node->param_values, exec_input_vals, node->output_values);
    }
    node->processed_state = PROCESSED;
}

void graph_compute_nodes(struct computation_graph *graph, int* node_ids, int n_nodes) {
    int i;
    for (i = 0; i < graph->n_nodes; i++) {
        graph->nodes[i].processed_state = UNPROCESSED;
    }
    for (i = 0; i < n_nodes; i++) {
    	int node_id = node_ids[i];
    	if (node_id < graph->n_nodes) {
    	    graph_compute_node_rec(graph, node_id, 0);
    	}
    }
}

int export_dot(const struct computation_graph* graph, FILE* of) {
    fprintf(of, "digraph G {\n"); // Header
    fprintf(of, "rankdir=\"LR\"\n"); // Horizontal layout

    // Draw all the nodes and their inputs
    int i;
    for (i = 0; i < graph->n_nodes; i++) {
        struct graph_node *node = &graph->nodes[i];
        // Create node
        fprintf(of, "\"%s\"[shape=record\nlabel=\"\n", node->name);
        fprintf(of, "<f0> %s\n", node->name); // Node name is port 0
        int j;
        // Create ports for inputs
        for (j = 0; j < node->type->n_inputs; j++) {
            fprintf(of, "|<f%d> --\\>%s\n", j+1, node->type->input_names[j]);
        }
        // Create ports for parameters
        for (j = 0; j < node->type->n_params; j++) {
            fprintf(of, "|<f%d> [%s=%.3f]\n", j+1+node->type->n_inputs, node->type->param_names[j],node->param_values[j]);
        }
        fprintf(of, "\"]\n"); // Close label bracket
        // Make connections from
        for (j = 0; j < node->type->n_inputs; j++) {
        	int input_id = node->input_srcs[j].controller_id;
        	if (input_id != -1) {
                struct graph_node* src_node = &graph->nodes[input_id];
                int output_id = node->input_srcs[j].controller_output;
                const char* output_name = src_node->type->output_names[output_id];
                fprintf(of, "\"%s\" -> \"%s\":f%d [label=\"%s=%.3f\"]\n", src_node->name, node->name, j+1, output_name, src_node->output_values[output_id]);
        	}
        }
    }
    fprintf(of, "}"); // Close graph
    return 0;
}
